/* fltetglass.cc - implementation of the tetris-glass widget.
 * Declarations of the shapes and the member functions.
 */
#include <time.h>
#include <string.h>

/* Define to switch off the pixmaps and use primitives instead: */
#undef SIMPLE_DESIGN

/* FLTK headers: */
#include <FL/Fl_Widget.H>
#include <FL/Fl_Pixmap.H>
#include <FL/fl_draw.H>

#ifndef SIMPLE_DESIGN

/* The pixmaps for drawing the glass: */
#include "square.xpm"
#include "left_border.xpm"
#include "right_border.xpm"
#include "bottom_border.xpm"
#include "top_border.xpm"
#include "lb_border.xpm"
#include "rb_border.xpm"
#include "lt_border.xpm"
#include "rt_border.xpm"

/* Image objects for the pixmaps: */
Fl_Pixmap square_img (square_xpm);
Fl_Pixmap left_border_img (left_border_xpm);
Fl_Pixmap right_border_img (right_border_xpm);
Fl_Pixmap bottom_border_img (bottom_border_xpm);
Fl_Pixmap top_border_img (top_border_xpm);
Fl_Pixmap lb_border_img (lb_border_xpm);
Fl_Pixmap rb_border_img (rb_border_xpm);
Fl_Pixmap lt_border_img (lt_border_xpm);
Fl_Pixmap rt_border_img (rt_border_xpm);

#else

#define TET_BASE_COLOR FL_BLUE
#define TET_AUX_COLOR FL_YELLOW

#endif // SIMPLE_DESIGN

/* Own header: */
#include "fltetglass.h"

/**** Functions and data of 'glass_widget': ****/

/* The constructor: */
glass_widget::glass_widget
 (
  int x, int y /* the left-upper corner of the widget */
 )
 : Fl_Widget (x, y,
               GLASS_WIDTH * SQUARE_DIM + 2 * BORDER_DIM,
                GLASS_HEIGHT * SQUARE_DIM + 2 * BORDER_DIM)
 {
  memset (contents, 0, GLASS_HEIGHT * GLASS_WIDTH * sizeof (char));
  random_type = old_time = 0;
  game_over_flag = true; /* initially the game is over */
  score = 0;
  shape_type = shape_rot = 0;
  cur_shape_x = cur_shape_y = 0;
 }

/* check_shape_position - checks the coordinates of the elements of
 * the given turned version of the current shape type placed at the
 * given position. If this shape can be placed on the given
 * place (described by the coordinates of the zero point of that shape)
 * this function returns 0, else it returns nonzero. The function
 * gets the following arguments:
 * i0, j0 - the glass coordinates to place the zero point of the current
 * shape at, rot - the given rotation of the current shape type (the
 * second index of the array 'shape'):
 */
bool glass_widget::check_shape_position (short i0, short j0, unsigned char rot)
 {
  struct shape_pos * element; /* the pointer to the array of the elements of the cur. shape */
  unsigned char len; /* the length of this array */
  unsigned char k;
  register short i, j;

  element = shape[shape_type][rot]->squares;
  len = shape[shape_type][rot]->num_of_squares;
  for (k = 0; k < len; k++)
   {
    i = i0 + element[k].x;
    j = j0 + element[k].y;
    if (i < 0 || i >= GLASS_WIDTH || j < 0)
     return 1;
    if (j < GLASS_HEIGHT && contents[j][i])
     return 1;
   };
  return 0;
 }

/* add_shape - adds the current shape to the contents of the glass
 * at the current position. It returns the increment of the score:
 */
int glass_widget::add_shape ()
 {
  struct shape_pos * element; /* the pointer to the array of the elements of the cur. shape */
  unsigned char len; /* the length of this array */
  unsigned char k;
  register unsigned short i, j;
  int overflown;

  element = shape[shape_type][shape_rot]->squares;
  len = shape[shape_type][shape_rot]->num_of_squares;
  overflown = 0;
  for (k = 0; k < len; k++)
   {
    i = cur_shape_x + element[k].x;
    j = cur_shape_y + element[k].y;
    if (j >= GLASS_HEIGHT)
     overflown++;
    else
     contents[j][i] = 1;
   };
  
  if (overflown)
    game_over_flag = true;

  return len - overflown;
 }

/* in_shape - verifies whether a given position belongs
 * to the shape and returns nonzero, if yes, or 0 otherwise:
 */
bool glass_widget::in_shape (short i0, short j0)
 {
  struct shape_pos * element; /* the pointer to the array of the elements of the cur. shape */
  unsigned char len; /* the length of this array */
  unsigned char k;

  element = shape[shape_type][shape_rot]->squares;
  len = shape[shape_type][shape_rot]->num_of_squares;
  i0 -= cur_shape_x; j0 -= cur_shape_y;
  if (j0 < element[0].y || j0 > element[len-1].y)
   return 0;
  for (k = 0; k < len; k++)
   if (element[k].x == i0 && element[k].y == j0)
    return 1;
  return 0;
 }

/* reduce_glass - removes the full lines from the contents
 * of the glass:
 */
void glass_widget::reduce_glass ()
 {
  register unsigned short i, j;

  j = 0;
  while (j < GLASS_HEIGHT)
   {
    /* Searching for the empty squares in the j-th line: */
    for (i = 0; i < GLASS_WIDTH; i++)
     if (contents[j][i] == 0)
      {
       j++; /* if found, go to the next line: */
       goto reduce_next_line;
      };
    /* Else shift the contents of the glass array: */
    if (j < GLASS_HEIGHT - 1)
     memcpy (&(contents[j][0]), &(contents[j+1][0]), GLASS_WIDTH * (GLASS_HEIGHT - 1 - j) * sizeof (char));
    /* and fill the the upper line with zeroes: */
    for (i = 0; i < GLASS_WIDTH; i++)
     contents[GLASS_HEIGHT-1][i] = 0;
    reduce_next_line: ;
   };
 }

/* launch_new_shape - initializes the new shape (i. e. chooses it
 * and makes it current). If the new shape cannot be put into the glass,
 * the function sets the variable 'game_over' to nonzero:
 */
void glass_widget::launch_new_shape ()
 {
  unsigned long t;

  t = time (NULL);
  random_type = (4013 * random_type + (t - old_time) + 1) % 5987;
  old_time = t;
  shape_type = random_type % NUM_OF_SHAPES;
  shape_rot = random_type % 4;
  cur_shape_x = shape[shape_type][shape_rot]->init_x;
  cur_shape_y = shape[shape_type][shape_rot]->init_y;
  
  if (check_shape_position (cur_shape_x, cur_shape_y, shape_rot))
   game_over_flag = true;
 }

/* reset_game - resets the game: */
void glass_widget::reset_glass ()
 {
  memset (contents, 0, GLASS_HEIGHT * GLASS_WIDTH * sizeof (char));
  game_over_flag = false;
  score = 0;
  do_callback ();
  launch_new_shape ();
  
  /* Redraw the glass: */
  redraw ();
 }

/* shift_shape - shift the current shape horizontally: */
void glass_widget::shift_shape
 (
  int shift /* the horizontal shift */
 )
 {
  /* Skip if the game is over: */
  if (game_over ()) return;
  
  if (! check_shape_position (cur_shape_x + shift, cur_shape_y, shape_rot))
   cur_shape_x += shift;
  
  /* Redraw the glass: */
  redraw ();
 }

/* turn_shape - rotates the current shape: */
void glass_widget::turn_shape ()
 {
  char t;
  
  /* Skip if the game is over: */
  if (game_over ()) return;
  
  t = (shape_rot + 1) % 4;
  if (! check_shape_position (cur_shape_x, cur_shape_y, t))
   shape_rot = t;
  
  /* Redraw the glass: */
  redraw ();
 }

/* drop_down - drop the current shape down
 * and add it to the glass contents:
 */
void glass_widget::drop_down ()
 {
  short t;
  
  /* Skip if the game is over: */
  if (game_over ()) return;
  
  /* Look for the lowest position of the shape: */
  t = cur_shape_y - 1;
  while (! check_shape_position (cur_shape_x, t, shape_rot))
   {
    cur_shape_y = t;
    t--;
   };
  score += add_shape (); /* add a shape to the glass contents */
  do_callback ();
  if (! game_over ())
   { /* if the glass has not been overflown */
    reduce_glass (); /* delete the empty lines */
    launch_new_shape (); /* launch the new shape */
   }
  else
   do_callback ();
  
  /* Redraw the glass: */
  redraw ();
 }

/* step - performs a step of the game: */
void glass_widget::step ()
 {
  /* Skip if the game is over: */
  if (game_over ()) return;
  
  /* Perform the game step: */
  if (check_shape_position (cur_shape_x, cur_shape_y - 1, shape_rot))
   {
    /* If the shape cannot fall farther, add it to the glass
     * contents and launch new shape or terminate the game
     */
    score += add_shape (); /* add a shape to the glass contents */
    do_callback ();
    if (! game_over ())
     { /* if the shape has not been overflown */
      reduce_glass (); /* delete the empty lines */
      launch_new_shape (); /* launch the new shape */
     }
    else
     do_callback ();
   }
  else /* else move the shape doun */
   cur_shape_y--;

  /* Redraw the glass: */
  redraw ();
 }

/* draw - the widget's draw method: */

#ifndef SIMPLE_DESIGN

void glass_widget::draw ()
 {
  unsigned i, j;
  unsigned dy;
  
  /* Plot the bottom: */
  dy = GLASS_HEIGHT * SQUARE_DIM + BORDER_DIM; /* the relative ordinate of the glass' bottom */
  lb_border_img.draw (x (), y () + dy);
  for (i = 0; i < GLASS_WIDTH; i++)
   bottom_border_img.draw (x () + SQUARE_DIM * i + BORDER_DIM, y () + dy);
  rb_border_img.draw (x () + SQUARE_DIM * GLASS_WIDTH + BORDER_DIM, y () + dy);
  
  /* Plot the glass rows. Every row consists of two border pixmaps
   * and the full or empty squares (according to the 'contents' array):
   */
  for (j = 0; j < GLASS_HEIGHT; j++)
   {
    dy = (GLASS_HEIGHT - j - 1) * SQUARE_DIM + BORDER_DIM;
    left_border_img.draw (0, y () + dy);
    for (i = 0; i < GLASS_WIDTH; i++)
     if (contents[j][i] || ((! game_over ()) && in_shape (i, j)))
	  /**/
      fl_draw_pixmap (square_xpm, x () + SQUARE_DIM * i + BORDER_DIM, y () + dy);
	  /**
	  square_img.draw (x () + SQUARE_DIM * i + BORDER_DIM, y () + dy);
	   **/
     else
      {
       fl_color (FL_BLACK);
       fl_rectf (x () + SQUARE_DIM * i + BORDER_DIM, y () + dy,
        SQUARE_DIM, SQUARE_DIM);
      };
    right_border_img.draw (x () + SQUARE_DIM * GLASS_WIDTH + BORDER_DIM, y () + dy);
   };
  
  /* Plot the top: */
  lt_border_img.draw (x (), y ());
  for (i = 0; i < GLASS_WIDTH; i++)
   top_border_img.draw (x () + SQUARE_DIM * i + BORDER_DIM, y ());
  rt_border_img.draw (x () + SQUARE_DIM * GLASS_WIDTH + BORDER_DIM, y ());
 }
 
#else

void glass_widget::draw ()
 {
  unsigned i, j;
  unsigned dy;
  
  /* Plot the walls: */
  dy = GLASS_HEIGHT * SQUARE_DIM + BORDER_DIM; /* the relative ordinate of the glass' bottom */
  fl_color (TET_BASE_COLOR);
  fl_rectf (x (), y () + dy,
   SQUARE_DIM * GLASS_WIDTH + 2 * BORDER_DIM, BORDER_DIM);
  fl_rectf (x (), y (),
   SQUARE_DIM * GLASS_WIDTH + 2 * BORDER_DIM, BORDER_DIM);
  fl_rectf (x (), y () + BORDER_DIM, BORDER_DIM, SQUARE_DIM * GLASS_HEIGHT);
  fl_rectf (x () + SQUARE_DIM * GLASS_WIDTH + BORDER_DIM, y () + BORDER_DIM,
   BORDER_DIM, SQUARE_DIM * GLASS_HEIGHT);
  
  /* Plot the glass rows. Every row consists of two border pixmaps
   * and the full or empty squares (according to the 'contents' array):
   */
  for (j = 0; j < GLASS_HEIGHT; j++)
   {
    dy = (GLASS_HEIGHT - j - 1) * SQUARE_DIM + BORDER_DIM;
    for (i = 0; i < GLASS_WIDTH; i++)
     if (contents[j][i] || ((! game_over ()) && in_shape (i, j)))
	  {
	   fl_color (TET_AUX_COLOR);
	   fl_rect (x () + SQUARE_DIM * i + BORDER_DIM, y () + dy,
        SQUARE_DIM, SQUARE_DIM);
	   fl_color (TET_BASE_COLOR);
	   fl_rectf (x () + SQUARE_DIM * i + BORDER_DIM + 1, y () + dy + 1,
        SQUARE_DIM - 2, SQUARE_DIM - 2);
	  }
     else
      {
       fl_color (FL_BLACK);
       fl_rectf (x () + SQUARE_DIM * i + BORDER_DIM, y () + dy,
        SQUARE_DIM, SQUARE_DIM);
      };
   };
 }

#endif

/**** The shapes: ****/

/* The descriptions of the shapes; every description consists of the
 * array of positions of the squares and the structure shape_descr; the
 * positions of the squares are in the acsent order of their ordinates:
 */
/* 1. **
 *    **
 */
struct shape_pos simple_square_elem []
 = {{0, 0}, {1, 0}, {0, 1}, {1, 1}};
struct shape_descr simple_square
 = {GLASS_WIDTH / 2, GLASS_HEIGHT - 1, simple_square_elem,
  sizeof (simple_square_elem) / sizeof (struct shape_pos)};

/* 2a. ****
 */
struct shape_pos line1_elem []
 = {{-2, 0}, {-1, 0}, {0, 0}, {1,0}};
struct shape_descr line1
 = {GLASS_WIDTH / 2, GLASS_HEIGHT - 1, line1_elem,
  sizeof (line1_elem) / sizeof (struct shape_pos)};

/* 2b. *
 *     *
 *     *
 *     *
 */
struct shape_pos line2_elem []
 = {{0, -1}, {0, 0}, {0, 1}, {0, 2}};
struct shape_descr line2
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, line2_elem,
  sizeof (line2_elem) / sizeof (struct shape_pos)};

/* 3a.  *
 *     ***
 */
struct shape_pos T1_elem []
 = {{-1, 0}, {0, 0}, {1, 0}, {0, 1}};
struct shape_descr T1
 = {GLASS_WIDTH / 2, GLASS_HEIGHT - 1, T1_elem,
  sizeof (T1_elem) / sizeof (struct shape_pos)};

/* 3b. *
 *     **
 *     *
 */
struct shape_pos T2_elem []
 = {{0, -1}, {0, 0}, {1, 0}, {0, 1}};
struct shape_descr T2
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, T2_elem,
  sizeof (T2_elem) / sizeof (struct shape_pos)};

/* 3c. ***
 *      *
 */
struct shape_pos T3_elem []
 = {{0, -1}, {-1, 0}, {0, 0}, {1, 0}};
struct shape_descr T3
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, T3_elem,
  sizeof (T3_elem) / sizeof (struct shape_pos)};

/* 3d. *
 *    **
 *     *
 */
struct shape_pos T4_elem []
 = {{0, -1}, {0, 0}, {-1, 0}, {0, 1}};
struct shape_descr T4
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, T4_elem,
  sizeof (T4_elem) / sizeof (struct shape_pos)};

/* 4a. *
 *     **
 *      *
 */
struct shape_pos S1_elem []
 = {{1, -1}, {0, 0}, {1, 0}, {0, 1}};
struct shape_descr S1
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, S1_elem,
  sizeof (S1_elem) / sizeof (struct shape_pos)};

/* 4b.  **
 *     **
 */
struct shape_pos S2_elem []
 = {{-1, -1}, {0, -1}, {0, 0}, {1, 0}};
struct shape_descr S2
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, S2_elem,
  sizeof (S2_elem) / sizeof (struct shape_pos)};

/* 5a.  *
 *     **
 *     *
 */
struct shape_pos Z1_elem []
 = {{0, -1}, {0, 0}, {1, 0}, {1, 1}};
struct shape_descr Z1
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, Z1_elem,
  sizeof (Z1_elem) / sizeof (struct shape_pos)};

/* 5b. **
 *      **
 */
struct shape_pos Z2_elem []
 = {{1, -1}, {0, -1}, {0, 0}, {-1, 0}};
struct shape_descr Z2
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, Z2_elem,
  sizeof (Z2_elem) / sizeof (struct shape_pos)};

/* 6a. *
 *     *
 *     **
 */
struct shape_pos L1_elem []
 = {{0, -1}, {1, -1}, {0, 0}, {0, 1}};
struct shape_descr L1
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, L1_elem,
  sizeof (L1_elem) / sizeof (struct shape_pos)};

/* 6b. ***
 *     *
 */
struct shape_pos L2_elem []
 = {{-1, -1}, {-1, 0}, {0, 0}, {1, 0}};
struct shape_descr L2
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, L2_elem,
  sizeof (L2_elem) / sizeof (struct shape_pos)};

/* 6c. **
 *      *
 *      *
 */
struct shape_pos L3_elem []
 = {{0, -1}, {0, 0}, {-1, 1}, {0, 1}};
struct shape_descr L3
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, L3_elem,
  sizeof (L3_elem) / sizeof (struct shape_pos)};

/* 6d.   *
 *     ***
 */
struct shape_pos L4_elem []
 = {{-1, 0}, {0, 0}, {1, 0}, {1, 1}};
struct shape_descr L4
 = {GLASS_WIDTH / 2, GLASS_HEIGHT - 1, L4_elem,
  sizeof (L4_elem) / sizeof (struct shape_pos)};

/* 7a. *
 *     *
 *    **
 */
struct shape_pos G1_elem []
 = {{-1, -1}, {0, -1}, {0, 0}, {0, 1}};
struct shape_descr G1
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, G1_elem,
  sizeof (G1_elem) / sizeof (struct shape_pos)};

/* 7b. *
 *     ***
 */
struct shape_pos G2_elem []
 = {{-1, 0}, {0, 0}, {1, 0}, {-1, 1}};
struct shape_descr G2
 = {GLASS_WIDTH / 2, GLASS_HEIGHT - 1, G2_elem,
  sizeof (G2_elem) / sizeof (struct shape_pos)};

/* 7c. **
 *     *
 *     *
 */
struct shape_pos G3_elem []
 = {{0, -1}, {0, 0}, {0, 1}, {1, 1}};
struct shape_descr G3
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, G3_elem,
  sizeof (G3_elem) / sizeof (struct shape_pos)};

/* 7d. ***
 *       *
 */
struct shape_pos G4_elem []
 = {{1, -1}, {-1, 0}, {0, 0}, {1, 0}};
struct shape_descr G4
 = {GLASS_WIDTH / 2, GLASS_HEIGHT, G4_elem,
  sizeof (G4_elem) / sizeof (struct shape_pos)};

/* The array of the pointers to the shape descriptions;
 * for every concrete i the variable shape[i] keeps four pointers
 * to the descriptions of the turned versions of one shape type
 * (i. e. that can be the same pointers for the simple square,
 * but for the shapes that are not invariant w.r.t. the rotations
 * these are different values):
 */
struct shape_descr * shape [NUM_OF_SHAPES] [4] =
 {
  {&simple_square, &simple_square, &simple_square, &simple_square},
  {&line1, &line2, &line1, &line2},
  {&T1, &T2, &T3, &T4},
  {&S1, &S2, &S1, &S2},
  {&Z1, &Z2, &Z1, &Z2},
  {&L1, &L2, &L3, &L4},
  {&G1, &G2, &G3, &G4}
 };

/* End of File */
