/* fltetglass.h - header of the tetris-glass widget.
 * Definition of the widget class and the auxiliary data types.
 */

/*** The sizes of the glass and its elements: ***/

/* The sizes of the glass (in 'squares'): */
#define GLASS_WIDTH 15
#define GLASS_HEIGHT 20

/* The size of the square: */
#define SQUARE_DIM 32

/* The width of the border (the border has everywhere the
 * same width, but sometimes it is its vertical and
 * sometimes horizontal size; the another dimention of the
 * border is SQUARE_DIM):
 */
#define BORDER_DIM 8

/* Sizes of the glass in pixels: */
#define GLASS_WIDTH_PIX (GLASS_WIDTH * SQUARE_DIM + 2 * BORDER_DIM)
#define GLASS_HEIGHT_PIX (GLASS_HEIGHT * SQUARE_DIM + 2 * BORDER_DIM)

/*** Structures and references for the game date: ***/

/* The shape square position structure: */
struct shape_pos
 {
  /* The relative coordinates: */
  short x;
  short y;
 };

/* The shape structure (for the descriptions of the shapes;
 * note that every turned version of the shape is a shape as well,
 * thus it must be declared too):
 */
struct shape_descr
 {
  /* The initial coordinates of the shape in the glass: */
  unsigned short init_x, init_y;
  /* The pointer to the array keeping the coordinates of
   * the squares (shape elements) of the shape relatively to
   * the shape's own zero point:
   */
  struct shape_pos * squares;
  /* The length of this array: */
  unsigned char num_of_squares;
 };

/* The number of classes of the shapes: */
#define NUM_OF_SHAPES 7

/* The array of the pointers to the shape descriptions: */
extern struct shape_descr * shape [NUM_OF_SHAPES] [4];

/*** The glass widget declaration: ***/

/* The widget provides the drawing of the tetris glass and the functions
 * to control the game. This widget does not have its own timer: The time
 * stepping should be done from the upper level widget. If the score
 * is changed or the glass is overflown (i.e. the game has become over),
 * the widget calls the callback.
 */

class glass_widget : public Fl_Widget
 {
  /* The array that keeps the contents of the glass
   * (except of the falling shape). The empty fields
   * are described by zeros and the filled ones
   * by the nonzero values:
   */
  char contents [GLASS_HEIGHT] [GLASS_WIDTH];

  /* The following two fields are used for generating a random shape: */
  unsigned long random_type, old_time;
  
  /* The widget's virtual functions: */
  void draw (void);

  /* The state of the game: nonzero iff the glass has been overflown: */
  bool game_over_flag;
  
  /* The score of the last game: */
  unsigned long score;

  unsigned char shape_type; /* cur. shape type (the cur. 1st index in 'shape') */
  unsigned char shape_rot; /* cur. shape rotation (the cur. 2nd index in 'shape') */
  /* Cur. coordinates of the cur. shape's zero point in the glass: */
  unsigned short cur_shape_x, cur_shape_y;
  
  /* Auxiliary functions: */
  
  bool check_shape_position (short, short, unsigned char);
  int add_shape (void);
  bool in_shape (short, short);
  void reduce_glass (void);
  void launch_new_shape (void);

  public:
  
  /* Widget's control methods: */
  
  bool game_over (void) /* returns non-zero iff the glass has been overflown */
   {
    return game_over_flag;
   };
  unsigned long game_score (void) /* return the current score of the game */
   {
    return score;
   };
  void reset_glass (void); /* resets the glass, starts a new game */
  void shift_shape /* shifts the shape horizontally */
   (
    int shift /* the horizontal shift */
   );
  void turn_shape (void); /* turns the shape */
  void drop_down (void); /* drops the shape down toward the bottom of the glass */
  void step (void); /* performs a step of the game */

  /* Constructor: */
  glass_widget
   (
    int x, int y /* the left-upper corner of the widget */
   );
 };

/* End of File */
