/* fltetris.cc - the main module of the fltk-tetris game.
 * Class declaration of the main window. Buttons and callbacks.
 * D. Logashenko
 */
#include <stdio.h>

/* FLTK headers: */
#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Button.H>
#include <FL/Enumerations.H>
#include <FL/forms.H>

/* The header of the tetris-glass widget: */
#include "fltetglass.h"

/* The time step of the game: */
#define GAME_PERIOD 0.5

/*** Positions and sizes of the widgets; fonts and colors: ***/

/* The coordinates of the left upper point of the glass: */
#define GLASS_LEFT_X 0
#define GLASS_TOP_Y 0

/* The dimention of a control button (all buttons are square and have
 * the same dimention):
 */
#define BUTTON_DIM 32
#define BUTTON_SIZES BUTTON_DIM, BUTTON_DIM

/* The size of the main window: */
#define MAIN_WND_WIDTH (GLASS_LEFT_X + GLASS_WIDTH_PIX + 2)
#define MAIN_WND_HEIGHT (GLASS_TOP_Y + GLASS_HEIGHT_PIX + 2 + BUTTON_DIM + 4)

/* The coordinates of the Start/Resume and Stop buttons: */
#define START_STOP_BUT_X (GLASS_LEFT_X + BORDER_DIM)
#define START_STOP_BUT_Y (GLASS_TOP_Y + GLASS_HEIGHT_PIX + 2)
#define START_STOP_BUT_COORD START_STOP_BUT_X, START_STOP_BUT_Y

/* The coordinates of the ShiftToRight button: */
#define SHIFT_TO_RIGHT_BUT_X (GLASS_LEFT_X + GLASS_WIDTH_PIX - BUTTON_DIM)
#define SHIFT_TO_RIGHT_BUT_Y START_STOP_BUT_Y
#define SHIFT_TO_RIGHT_BUT_COORD SHIFT_TO_RIGHT_BUT_X, SHIFT_TO_RIGHT_BUT_Y

/* The coordinates of the ShiftToLeft button: */
#define SHIFT_TO_LEFT_BUT_X (SHIFT_TO_RIGHT_BUT_X - BUTTON_DIM - 3)
#define SHIFT_TO_LEFT_BUT_Y START_STOP_BUT_Y
#define SHIFT_TO_LEFT_BUT_COORD SHIFT_TO_LEFT_BUT_X, SHIFT_TO_LEFT_BUT_Y

/* The coordinates of the Turn button: */
#define TURN_BUT_X (SHIFT_TO_LEFT_BUT_X - BUTTON_DIM - 10)
#define TURN_BUT_Y START_STOP_BUT_Y
#define TURN_BUT_COORD TURN_BUT_X, TURN_BUT_Y

/* The coordinates of the Drop button: */
#define DROP_BUT_X (TURN_BUT_X - BUTTON_DIM - 10)
#define DROP_BUT_Y START_STOP_BUT_Y
#define DROP_BUT_COORD DROP_BUT_X, DROP_BUT_Y

/* The coordinates and the sizes of the indicator: */
#define TABLEAU_X START_STOP_BUT_X + BUTTON_DIM + 10
#define TABLEAU_Y START_STOP_BUT_Y
#define TABLEAU_W BUTTON_DIM * 5
#define TABLEAU_H BUTTON_DIM
#define TABLEAU_FSIZE 30

/*** The special button class: ***/

/* Buttons of this class do not have the keyboard focus. */

class tetris_button : public Fl_Button
 {
  public:
  /* The constructor: */
  tetris_button (int x, int y, int w, int h, ulong key, const char * label = 0)
   : Fl_Button (x, y, w, h, label)
   {
    shortcut (key);
   };
  
  private:
  /* The event handler: */
  int handle (int event)
   {
    if (event == FL_FOCUS)
     return 0; /* no focus */
    return Fl_Button::handle (event);
   };
 };

/*** The score indicator class: ***/

class tetris_tableau : public Fl_Widget
 {
  char tableau_str [32];
  
  public:
  
  void value /* sets the value of the score */
   (
    unsigned long score /* the score to set */
   )
   {
    snprintf (tableau_str, sizeof (tableau_str), "%lu", score);
    redraw ();
   };
  
  /* The constructor: */
  tetris_tableau (int x, int y, int w, int h)
   : Fl_Widget (x, y, w, h, NULL)
   {
    tableau_str [0] = '\0';
   };
  
  private:
  
  /* The drawing routine: */
  void draw (void)
   {
    /* Draw the box */
    fl_draw_box (FL_DOWN_BOX, x (), y (), w (), h (), FL_GRAY);
    
    /* Draw the string: */
    fl_font (FL_BOLD, TABLEAU_FSIZE);
    fl_color (FL_BLACK);
    fl_draw (tableau_str, x (), y (), w (), h (), FL_ALIGN_CENTER);
   };
 };

/*** The main window widget class: ***/

class tetris_window : public Fl_Double_Window
 {
  char pause_phase; /* to blink with the pause-button */
  
  /* The "Game Paused" flag: nonzero iff the game is paused by the user */
  bool game_paused;
  
  /* Child widgets: */
  tetris_button start_stop_btn; /* the Start/Pause button */
  tetris_button shift_left_btn; /* the Shift-Left button */
  tetris_button shift_right_btn; /* the Shift-Right button */
  tetris_button turn_btn; /* the rotation button */
  tetris_button drop_btn; /* the Drop-Down button */
  glass_widget glass; /* the 'glass' of the tetris game */
  tetris_tableau tableau; /* the score indicator */
  
  /* Callbacks of the widgets: */
  static void start_stop_cb (Fl_Widget * w, void * p);
  static void shift_left_cb (Fl_Widget * w, void * p);
  static void shift_right_cb (Fl_Widget * w, void * p);
  static void turn_cb (Fl_Widget * w, void * p);
  static void drop_down_cb (Fl_Widget * w, void * p);
  static void glass_cb (Fl_Widget * w, void * p);

  /* The functions for controlling the game: */
  
  inline void start_stop (void); /* starts/pauses/stops the game */
  
  public:
  
  inline void time_step (void); /* performs one time step of the game and blinking */
  
  /* The constructor: */
  inline tetris_window ();
 };

/*** The functions controlling the game: ***/

/* start_stop - starts the game, pauses it or continues it if paused: */
void tetris_window::start_stop ()
 {
  if (glass.game_over ()) /* if the game has not been started at all */
   {
    /* Change the button label: */
    start_stop_btn.label ("@||");
    /* Reset the game: */
    glass.reset_glass ();
    game_paused = false;
   }
  else if (game_paused) /* if the game has been paused */
   {
    /* Change the button pixmap: */
    start_stop_btn.label ("@||");
    /* Cancel the pause: */
    game_paused = false;
   }
  else /* if the game is going */
   {
    /* Change the button pixmap: */
    start_stop_btn.label ("@UpArrow");
    pause_phase = 0;
    /* Set the pause flag: */
    game_paused = true;
   };
 }

/* time_step - one time step of the game: */
void tetris_window::time_step ()
 {
  /* If the game is over or paused, do nothing: */
  if (! glass.game_over ())
   if (game_paused)
    {
     /* Blink on the button: */
     pause_phase = (pause_phase + 1) % 2;
     start_stop_btn.label ((pause_phase == 0)? "@UpArrow" : "@DnArrow");
    }
   else
    glass.step ();
 }

/*** The callback functions of the child widgets: ***/

void tetris_window::glass_cb (Fl_Widget * w, void * p)
 {
  tetris_window * tetris = (tetris_window *) p;
  if (tetris->glass.game_over ())
   tetris->start_stop_btn.label ("@UpArrow");
  tetris->tableau.value (tetris->glass.game_score ());
 }

void tetris_window::start_stop_cb (Fl_Widget * w, void * p)
 {
  ((tetris_window *) p)->start_stop ();
 }

void tetris_window::drop_down_cb (Fl_Widget * w, void * p)
 {
  ((tetris_window *) p)->glass.drop_down ();
 }

void tetris_window::turn_cb (Fl_Widget * w, void * p)
 {
  ((tetris_window *) p)->glass.turn_shape ();
 }

void tetris_window::shift_left_cb (Fl_Widget * w, void * p)
 {
  ((tetris_window *) p)->glass.shift_shape (-1);
 }

void tetris_window::shift_right_cb (Fl_Widget * w, void * p)
 {
  ((tetris_window *) p)->glass.shift_shape (1);
 }

/*** The timer handler: ***/
void timer_handler (void * p)
 {
  /* Call the time stepping routine: */
  ((tetris_window *) p)->time_step ();
  
  /* Reset the timer: */
  Fl::add_timeout (GAME_PERIOD, timer_handler, p);
 }

/*** The constructor of the 'tetris_window' class: ***/
tetris_window::tetris_window ()
 : Fl_Double_Window (MAIN_WND_WIDTH, MAIN_WND_HEIGHT, "FLTetris"),
   start_stop_btn (START_STOP_BUT_COORD, BUTTON_SIZES, FL_Enter, "@UpArrow"),
   shift_left_btn (SHIFT_TO_LEFT_BUT_COORD, BUTTON_SIZES, FL_Left, "@<-"),
   shift_right_btn (SHIFT_TO_RIGHT_BUT_COORD, BUTTON_SIZES, FL_Right, "@->"),
   turn_btn (TURN_BUT_COORD, BUTTON_SIZES, FL_Up, "@reload"),
   drop_btn (DROP_BUT_COORD, BUTTON_SIZES, ' ', "@2->"),
   glass (GLASS_LEFT_X, GLASS_TOP_Y),
   tableau (TABLEAU_X, TABLEAU_Y, TABLEAU_W, TABLEAU_H)
 {
  /* No further children: */
  end ();
  
  /* Set the game variables: */
  game_paused = false; /* initially the game is not paused */
  pause_phase = 0;
  
  /* Set the callback functions: */
  start_stop_btn.callback (start_stop_cb, this);
  drop_btn.callback (drop_down_cb, this);
  turn_btn.callback (turn_cb, this);
  shift_left_btn.callback (shift_left_cb, this);
  shift_right_btn.callback (shift_right_cb, this);
  glass.callback (glass_cb, this);
 }

/*** The main function: ***/
int main (int argc, char * * argv)
 {
  tetris_window window;
  window.show (argc, argv);
  Fl::add_timeout (GAME_PERIOD, timer_handler, &window);
  return Fl::run ();
 }

/* End of File */
